package ru.dbotthepony.mc.otm.storage.powered

import it.unimi.dsi.fastutil.HashCommon.mix
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectLinkedOpenHashSet
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.storage.IStorageEventConsumer
import ru.dbotthepony.mc.otm.storage.IStorageProvider
import ru.dbotthepony.mc.otm.storage.StorageStack
import java.math.BigInteger
import java.util.*
import java.util.function.Supplier

class PoweredStorageProvider<T : StorageStack<T>>(val parent: IStorageProvider<T>, val energy: Supplier<IMatteryEnergyStorage>) : IStorageProvider<T> by parent {
	override fun equals(other: Any?): Boolean {
		return other is PoweredStorageProvider<*> && other.parent == parent
	}

	override fun hashCode(): Int {
		return parent.hashCode()
	}

	override fun toString(): String {
		return "PoweredStorageProvider[$parent]"
	}

	private class Proxy<T : StorageStack<T>>(private val parent: IStorageEventConsumer<T>, private val powered: PoweredStorageProvider<T>) : IStorageEventConsumer<T> {
		override val storageType: StorageStack.Type<T>
			get() = parent.storageType

		override fun onStackAdded(stack: T, id: UUID, provider: IStorageProvider<T>) {
			parent.onStackAdded(stack, id, powered)
		}

		override fun onStackChanged(stack: T, id: UUID) {
			parent.onStackChanged(stack, id)
		}

		override fun onStackRemoved(id: UUID) {
			parent.onStackRemoved(id)
		}

		override fun toString(): String {
			return "PoweredStorageProvider.Proxy[$parent with $powered]"
		}

		override fun equals(other: Any?): Boolean {
			return other is Proxy<*> && parent == other.parent && powered == other.powered
		}

		override fun hashCode(): Int {
			return mix(parent.hashCode() * 31 + powered.hashCode())
		}
	}

	override fun addListener(listener: IStorageEventConsumer<T>): Boolean {
		return parent.addListener(Proxy(listener, this))
	}

	override fun removeListener(listener: IStorageEventConsumer<T>): Boolean {
		return parent.removeListener(Proxy(listener, this))
	}

	override fun extractStack(id: UUID, amount: BigInteger, simulate: Boolean): T {
		val extracted = parent.extractStack(id, amount, true)
		if (extracted.isEmpty) return extracted

		val energy = energy.get()
		val required = storageType.energyPerExtract(extracted)

		if (energy.extractEnergy(required, true) == required) {
			if (!simulate) {
				val extracted2 = parent.extractStack(id, amount, false)

				if (extracted.count != extracted2.count) {
					energy.extractEnergy(storageType.energyPerExtract(extracted2), false)
				} else {
					energy.extractEnergy(required, false)
				}

				return extracted2
			}

			return extracted
		}

		return storageType.empty
	}
}
